import React, {Component} from 'react'
import PropTypes from 'prop-types'
import { is, fromJS } from 'immutable'
import { notification, Modal } from 'antd'

import Api from '@/api'
import UtilsDM from '@/utils/utils-datamanage.js'
import asyncComponent from '@/utils/asyncComponent'
import MKEmitter from '@/utils/events.js'
import './index.scss'

// 通用组件
const MainSearch = asyncComponent(() => import('@/tabviews/zshare/topSearch'))
const MainAction = asyncComponent(() => import('@/tabviews/zshare/actionList'))
const MainTable = asyncComponent(() => import('@/tabviews/custom/components/share/normalTable'))

class MkBaseTable extends Component {
  static propTpyes = {
    config: PropTypes.object,        // 组件配置信息
  }

  state = {
    BID: '',              // 上级ID
    BData: '',            // 上级组件行数据
    config: {},           // 页面配置信息，包括按钮、搜索、显示列、标签等
    actions: null,        // 按钮集
    columns: null,        // 显示列
    setting: null,        // 页面全局设置：数据源、按钮及显示列固定、主键等
    data: [],             // 列表数据集
    selectedData: [],     // 已选表格数据
    total: 0,             // 总数
    loading: false,       // 列表数据加载中
    pageIndex: 1,         // 页码
    pageSize: 10,         // 每页数据条数
    orderBy: '',          // 排序
    search: '',           // 搜索条件数组，使用时需分场景处理
    statFValue: []        // 合计值
  }

  /**
   * @description 初始化处理
   */
  UNSAFE_componentWillMount () {
    let _config = fromJS(this.props.config).toJS()

    let BID = ''
    let BData = ''

    if (_config.setting.supModule) {
      BData = window.GLOB.CacheData.get(_config.setting.supModule)
    } else {
      BData = window.GLOB.CacheData.get(_config.$pageId)
    }
    if (BData) {
      BID = BData.$BID || ''
    }

    if (_config.wrap.controlField) {
      if (_config.wrap.controlVal) {
        _config.wrap.controlVal = _config.wrap.controlVal.split(',')
      } else {
        _config.wrap.controlVal = ['']
      }
    }

    let setting = {..._config.setting, ..._config.wrap}

    if (setting.selected !== 'always' && setting.selected !== 'init' && setting.selected !== 'sign') {
      setting.selected = 'false'
    } else {
      setting.orisel = true
    }

    _config.style = _config.style || {}

    this.setState({
      pageSize: setting.pageSize || 10,
      BID: BID || '',
      BData: BData || '',
      config: _config,
      setting: setting,
      actions: _config.action,
      columns: _config.cols,
      search: _config.$searches
    }, () => {
      if (_config.setting.onload === 'true') {
        setTimeout(() => {
          this.loadmaindata()
          this.getStatFieldsValue()
        }, _config.setting.delay || 0)
      }
    })
  }

  /**
   * @description 主表数据加载
   * @param { Boolean } reset  表格是否重置
   * @param { String }  repage 表格是否重置页码
   */
  async loadmaindata (reset, repage, id) {
    const { setting, config, search, orderBy, BID, pageIndex, pageSize, BData } = this.state

    if (setting.supModule && !BID) { // BID 不存在时，不做查询
      this.setState({
        data: [],
        selectedData: [],
        total: 0
      })
      
      MKEmitter.emit('resetSelectLine', config.uuid, '', '') // 广播数据切换
      reset && MKEmitter.emit('resetTable', config.uuid, 'true') // 列表重置
      if (setting.$hasSyncModule) {
        MKEmitter.emit('syncBalconyData', config.uuid, [], false)
      }
      return
    }

    let searches = fromJS(search).toJS()
    if (config.setting.useMSearch) { // 主表搜索条件
      let mainSearch = window.GLOB.SearchBox.get(config.$searchId) || []
      let keys = config.$s_keys || []
      mainSearch.forEach(item => {
        if (keys.includes(item.key.toLowerCase())) return

        searches.push(item)
      })
    }

    if (config.$s_req && searches.filter(item => item.required && item.value === '').length > 0) {
      return
    }

    this.setState({
      loading: true
    })

    let _orderBy = orderBy || setting.order
    let param = UtilsDM.getQueryDataParams(setting, searches, _orderBy, pageIndex, pageSize, BID)

    let result = await Api.genericInterface(param)
    if (result.status) {
      if (repage === 'false' && result.data && result.data.length === 0 && result.total > 0 && pageIndex > 1) {
        let _pageIndex = Math.ceil(result.total / pageSize)

        if (_pageIndex < pageIndex) {
          MKEmitter.emit('resetTable', config.uuid, 'repage', _pageIndex)
          this.setState({
            pageIndex: _pageIndex,
            data: [],
            selectedData: [],
            total: result.total
          }, () => {
            this.loadmaindata()
          })
          return
        }
      }

      if ((setting.selected !== 'false' || (setting.orisel && id)) && result.data && result.data.length > 0) {
        setTimeout(() => {
          MKEmitter.emit('mkCheckTopLine', config.uuid, id, setting.selected)
        }, 200)
        if (setting.selected === 'init') {
          this.setState({setting: {...setting, selected: 'false'}})
        }
      } else {
        MKEmitter.emit('resetSelectLine', config.uuid, '', '') // 广播数据切换
        if (setting.$hasSyncModule) {
          MKEmitter.emit('syncBalconyData', config.uuid, [], false)
        }
      }
      
      reset && MKEmitter.emit('resetTable', config.uuid, repage) // 列表重置

      let start = 1
      if (setting.laypage) {
        start = pageSize * (pageIndex - 1) + 1
      }

      let data = result.data.map((item, index) => {
        item.key = index
        item.$$uuid = item[setting.primaryKey] || ''
        item.$$key = '' + item.key + item.$$uuid
        item.$$BID = BID || ''
        item.$$BData = BData || ''
        item.$Index = start + index + ''

        if (config.absFields) {
          config.absFields.forEach(f => {
            if (!isNaN(item[f])) {
              item[f] = Math.abs(item[f])
            }
          })
        }
        if (setting.controlField) {
          if (setting.controlVal.includes(item[setting.controlField] + '')) {
            item.$disabled = true
          }
        }
        
        return item
      })

      let total = result.total || 0
      if (config.setting.custompage && data.length) {
        total = data[data.length - 1].mk_total || 0
      }

      this.setState({
        data: data,
        selectedData: [],
        total: total,
        loading: false
      })

      if (config.autoMatic) {
        if (result.data && result.data.length > 0) {
          MKEmitter.emit('autoGetData', config.MenuID)
        } else {
          MKEmitter.emit('autoMaticOver', config.MenuID)
        }
      }

      if (result.message) {
        if (result.ErrCode === 'Y') {
          Modal.success({
            title: result.message
          })
        } else if (result.ErrCode === 'S') {
          notification.success({
            top: 92,
            message: result.message,
            duration: 2
          })
        }
      }
    } else {
      this.setState({
        loading: false
      })

      if (config.autoMatic) {
        MKEmitter.emit('autoMaticError', config.MenuID)
      }
      
      if (!result.message) return
      if (result.ErrCode === 'N') {
        Modal.error({
          title: result.message,
        })
      } else if (result.ErrCode !== '-2') {
        notification.error({
          top: 92,
          message: result.message,
          duration: 10
        })
      }
    }
  }

  /**
   * @description 获取单行数据
   */ 
  async loadmainLinedata (id) {
    const { setting, config, search, orderBy, BID, pageIndex, pageSize, BData } = this.state

    let searches = fromJS(search).toJS()
    if (config.setting.useMSearch) { // 主表搜索条件
      let mainSearch = window.GLOB.SearchBox.get(config.$searchId) || []
      let keys = config.$s_keys || []
      mainSearch.forEach(item => {
        if (keys.includes(item.key.toLowerCase())) return

        searches.push(item)
      })
    }

    this.setState({
      loading: true
    })

    let _orderBy = orderBy || setting.order
    let param = UtilsDM.getQueryDataParams(setting, searches, _orderBy, pageIndex, pageSize, BID, id)

    let result = await Api.genericInterface(param)
    if (result.status) {
      let data = fromJS(this.state.data).toJS()
      let selectedData = fromJS(this.state.selectedData).toJS()

      if (result.data && result.data[0]) {
        let _data = result.data[0] || {}
        _data.$$uuid = _data[setting.primaryKey] || ''
        _data.$$BID = BID || ''
        _data.$$BData = BData || ''

        if (config.absFields) {
          config.absFields.forEach(f => {
            if (!isNaN(_data[f])) {
              _data[f] = Math.abs(_data[f])
            }
          })
        }

        data = data.map(item => {
          if (item.$$uuid === _data.$$uuid) {
            _data.key = item.key
            _data.$$key = '' + item.key + item.$$uuid
            _data.$Index = item.$Index
            return _data
          } else {
            return item
          }
        })

        if (!_data.$Index) {
          this.setState({
            loading: false
          })
          return
        }

        selectedData = selectedData.map(item => {
          if (_data.$$uuid === item.$$uuid) {
            return _data
          }
          return item
        })

        if (selectedData.length && _data.$$uuid === selectedData[selectedData.length - 1].$$uuid) {
          MKEmitter.emit('resetSelectLine', config.uuid, _data.$$uuid || '', _data)
        }
        
        this.setState({
          data,
          selectedData,
          loading: false
        })
      } else {
        let index = data.findIndex(item => item.$$uuid === id)

        if (index === -1) {
          this.setState({
            loading: false
          })
          return
        }

        data = data.filter(item => item.$$uuid !== id)
        data = data.map((item, i) => {
          item.key = i
          item.$$key = '' + item.key + item.$$uuid

          return item
        })

        this.setState({
          data: data,
          loading: false
        })

        setTimeout(() => {
          MKEmitter.emit('resetTable', config.uuid, 'delete', index)
        }, 20)
      }
    } else {
      this.setState({
        loading: false
      })
      notification.error({
        top: 92,
        message: result.message,
        duration: 10
      })
    }
  }

  /**
   * @description 获取合计字段值
   */
  getStatFieldsValue = () => {
    const { setting, config, search, BID, orderBy } = this.state

    if (setting.supModule && !BID) { // BID 不存在时，不做查询
      this.setState({
        statFValue: []
      })
      return
    }

    if (config.statFields.length === 0 || setting.interType !== 'system' || !setting.dataresource) return

    let searches = fromJS(search).toJS()
    if (config.setting.useMSearch) { // 主表搜索条件
      let mainSearch = window.GLOB.SearchBox.get(config.$searchId) || []
      let keys = config.$s_keys || []
      mainSearch.forEach(item => {
        if (keys.includes(item.key.toLowerCase())) return

        searches.push(item)
      })
    }

    if (config.$s_req && searches.filter(item => item.required && item.value === '').length > 0) {
      return
    }

    let _orderBy = orderBy || setting.order
    let param = UtilsDM.getStatQueryDataParams(setting, config.statFields, searches, _orderBy, BID)

    Api.genericInterface(param).then(res => {
      if (res.status) {
        let _data = res.data[0]
        let values = []

        if (_data) {
          config.statFields.forEach(item => {
            if (_data[item.field] || _data[item.field] === 0) {
              let val = +_data[item.field]
              if (isNaN(val)) {
                val = 0
              }
              val = val.toFixed(item.decimal)
              values.push({label: item.label, value: val})
            }
          })
        }
        this.setState({
          statFValue: values
        })
      } else {
        this.setState({
          statFValue: []
        })
        notification.error({
          top: 92,
          message: res.message,
          duration: 10
        })
      }
    })
  }

  /**
   * @description 搜索条件改变时，重置表格数据
   * 含有初始不加载的页面，修改设置
   */
  refreshbysearch = (searches) => {
    const { config} = this.state

    this.setState({
      pageIndex: 1,
      search: searches
    }, () => {
      this.loadmaindata(true, 'true')
      this.getStatFieldsValue()
    })

    if (config.$main) {
      window.GLOB.SearchBox.set(config.$searchId, searches)
      MKEmitter.emit('searchRefresh', config.$searchId)
    }
  }

  /**
   * @description 表格条件改变时重置数据（分页或排序）
   */
  refreshbytable = (pagination, filters, sorter) => {
    if (!sorter) { // 无人值守
      this.setState({
        pageIndex: pagination.pageIndex
      }, () => {
        this.loadmaindata()
      })
    } else {
      if (sorter.order) {
        let _chg = {
          ascend: 'asc',
          descend: 'desc'
        }
        sorter.order = _chg[sorter.order]
      }
  
      this.setState({
        pageIndex: pagination.current,
        pageSize: pagination.pageSize,
        orderBy: (sorter.field && sorter.order) ? `${sorter.field} ${sorter.order}` : ''
      }, () => {
        this.loadmaindata()
      })
    }
  }

  /**
   * @description 表格刷新
   */
  reloadtable = (btn, id) => {
    if (!btn || btn.resetPageIndex !== 'false') {
      this.setState({
        pageIndex: 1
      }, () => {
        this.loadmaindata(true, 'true', id)
        this.getStatFieldsValue()
      })
    } else {
      this.loadmaindata(true, 'false', id)
      this.getStatFieldsValue()
    }
  }

  /**
   * @description 导出Excel时，获取页面搜索排序等参数
   */
  queryModuleParam = (menuId, callback) => {
    const { config, orderBy, search, setting } = this.state

    if (config.uuid !== menuId) return

    let searches = search ? fromJS(search).toJS() : []
    if (config.setting.useMSearch) { // 主表搜索条件
      let mainSearch = window.GLOB.SearchBox.get(config.$searchId) || []
      let keys = config.$s_keys || []
      mainSearch.forEach(item => {
        if (keys.includes(item.key.toLowerCase())) return

        searches.push(item)
      })
    }

    callback({
      orderBy: orderBy || setting.order,
      search: searches
    })
  }

  reloadData = (menuId, id) => {
    const { config } = this.state

    if (config.uuid !== menuId) return

    if (!id) {
      this.reloadtable()
    } else {
      this.loadmainLinedata(id)
    }
  }

  resetParentParam = (MenuID, id, data) => {
    const { setting } = this.state

    if (!setting.supModule || setting.supModule !== MenuID) return
    if (id !== this.state.BID || id !== '') {
      this.setState({
        pageIndex: 1,
        BID: id,
        BData: data
      }, () => {
        this.loadmaindata(true, 'true')
        this.getStatFieldsValue()
      })
    }
  }

  /**
   * @description 按钮执行完成后页面刷新
   * @param {*} menuId     // 菜单Id
   * @param {*} position   // 刷新位置
   * @param {*} btn        // 执行的按钮
   */
  refreshByButtonResult = (menuId, position, btn, id, lines) => {
    const { config, BID } = this.state

    if (config.uuid !== menuId) return

    if (position === 'line') {
      if (lines && lines.length === 1) {
        this.loadmainLinedata(lines[0].$$uuid)
      } else {
        this.reloadtable(btn, id)
      }
    } else if ((position === 'mainline' || position === 'popclose') && config.setting.supModule && BID) { // 刷新源组件时，附带刷新上级行与当前组件
      MKEmitter.emit('reloadData', config.setting.supModule, BID)
    } else {
      this.reloadtable(btn, id)
    }
  }

  searchRefresh = (searchId) => {
    const { config } = this.state

    if (config.$searchId !== searchId) return
    
    this.setState({pageIndex: 1}, () => {
      this.reloadtable()
    })
  }

  shouldComponentUpdate (nextProps, nextState) {
    return !is(fromJS(this.props), fromJS(nextProps)) || !is(fromJS(this.state), fromJS(nextState))
  }

  componentDidMount () {
    const { config } = this.state

    MKEmitter.addListener('reloadData', this.reloadData)
    MKEmitter.addListener('resetSelectLine', this.resetParentParam)
    MKEmitter.addListener('queryModuleParam', this.queryModuleParam)
    MKEmitter.addListener('refreshByButtonResult', this.refreshByButtonResult)

    if (config.setting.useMSearch) {
      MKEmitter.addListener('searchRefresh', this.searchRefresh)
    }
  }

  /**
   * @description 组件销毁，清除state更新，清除快捷键设置
   */
  componentWillUnmount () {
    this.setState = () => {
      return
    }
    MKEmitter.removeListener('reloadData', this.reloadData)
    MKEmitter.removeListener('searchRefresh', this.searchRefresh)
    MKEmitter.removeListener('resetSelectLine', this.resetParentParam)
    MKEmitter.removeListener('queryModuleParam', this.queryModuleParam)
    MKEmitter.removeListener('refreshByButtonResult', this.refreshByButtonResult)
  }

  render() {
    const { BID, setting, actions, config, columns, selectedData, BData, data } = this.state
    
    return (
      <div className="custom-base-table" style={config.style}>
        {config.search.length ?
          <MainSearch BID={BID} config={config} refreshdata={this.refreshbysearch}/> : null
        }
        {actions.length > 0 ? <MainAction
          BID={BID}
          setting={setting}
          actions={actions}
          BData={BData}
          columns={config.columns}
          selectedData={selectedData}
        /> : <div style={{height: '25px'}}></div>}
        <div className="main-table-box">
          <MainTable
            data={data}
            setting={setting}
            columns={columns}
            MenuID={config.uuid}
            fields={config.columns}
            total={this.state.total}
            autoMatic={config.autoMatic}
            lineMarks={config.lineMarks}
            loading={this.state.loading}
            refreshdata={this.refreshbytable}
            statFValue={this.state.statFValue}
            chgSelectData={(selects) => this.setState({selectedData: selects})}
          />
        </div>
      </div>
    )
  }
}

export default MkBaseTable